#! /usr/bin/env python
# -*- coding: utf-8 -*-
# >>
#     gevent-tasks, 2017
# <<

__author__ = 'Blake VandeMerwe'
__version__ = '0.1.0'
__license__ = 'MIT'
__contact__ = 'blakev@null.net'
__url__ = 'https://github.com/blakev/gevent-tasks'

from gevent_tasks.manager import Task, TaskManager, TaskPool

__all__ = [
    'Task',
    'TaskManager',
    'TaskPool'
]



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# >>
#     gevent-tasks, 2017
# <<

import math
from logging import getLogger
from collections import OrderedDict

from gevent.pool import Pool

from gevent_tasks.tasks import Task, TaskPool


class TaskManager(object):
    def __init__(self, pool=None, logger=None, *tasks):
        # type: (Pool, *Task) -> self

        size = None

        if isinstance(pool, int):
            size = pool

        if pool is None:
            if tasks:
                size = math.ceil(len(tasks) * 0.75)
            else:
                size = 5

        if size is not None:
            pool = TaskPool(size=size)

        self._pool = pool
        self._tasks = OrderedDict()
        self.logger = logger or getLogger('%s.TaskManager' % __name__)

        for task in tasks:
            self.add(task)

    def __repr__(self):
        return '<TaskManager(tasks=%d,capacity=%d)>' % (
            len(self._tasks), self._pool.size)

    def __iter__(self):
        for task in self._tasks.values():
            yield task

    @property
    def pool(self):
        return self._pool

    @property
    def task_names(self):
        return [t for t in self._tasks.keys()]

    def add(self, task, start=False):
        # type: (Task, bool) -> None
        if task.name in self._tasks:
            raise KeyError(task.name)
        task.pool = self._pool
        self._tasks[task.name] = task
        if start and not task.is_running:
            task.start()

    def start(self, task_name):
        # type: (str) -> None
        t = self._tasks.get(task_name, None)
        if t:
            t.start()

    def start_all(self):
        # type: () -> None
        for task in self.task_names:
            self.start(task)

    def stop(self, task_name, force=False):
        # type: (str, bool) -> None
        t = self._tasks.get(task_name, None)
        if t:
            t.stop(force)

    def stop_all(self, force=False):
        # type: (bool) -> None
        for task in self.task_names:
            self.stop(task, force)

    def remove_task(self, task, force=False):
        if hasattr(task, 'name'):
            name = task.name
        else:
            name = task
        t = self._tasks.pop(name, None)
        if t:
            t.stop(force)

    def remove_all(self, force=False):
        # type: (bool) -> None
        for task in self.task_names:
            self.remove_task(task, force)





#! /usr/bin/env python
# -*- coding: utf-8 -*-
# >>
#   gevent-tasks, 2017
#
#       Background task management using gevent green threads.
#
#   Blake VandeMerwe
#   blakev@null.net
# <<

import os
import re
import sys
from os.path import dirname, join as pjoin
from setuptools import setup, find_packages, Command
from setuptools.command.test import test as TestCommand

ATTRIBUTES = re.compile(r"^__(\w+)__\s+=\s+'(.*?)'$", re.S + re.M)

dname = dirname(__file__)

with open(pjoin(dname, 'gevent_tasks', '__init__.py')) as ins_file:
    text = ins_file.read()
    attrs = ATTRIBUTES.findall(text)
    A = dict(attrs)

with open(pjoin(dname, 'requirements.txt')) as ins_file:
    requirements = map(str.strip, ins_file.readlines())


class RunTests(TestCommand):
    """
    Run the unit tests.
    By default, `python setup.py test` fails if tests/ isn't a Python
    module (that is, if the tests/ directory doesn't contain an
    __init__.py file). But the tests/ directory shouldn't contain an
    __init__.py file and tests/ shouldn't be a Python module. See
      http://doc.pytest.org/en/latest/goodpractices.html
    Running the unit tests manually here enables `python setup.py test`
    without tests/ being a Python module.
    """
    def run_tests(self):
        from unittest import TestLoader, TextTestRunner
        tests_dir = pjoin(dirname(__file__), 'tests')
        suite = TestLoader().discover(tests_dir)
        result = TextTestRunner().run(suite)
        sys.exit(0 if result.wasSuccessful() else -1)


long_description = (
    'Information and documentation found can be found'
    ' at https://github.com/blakev/gevent-tasks.')


setup(
    name='gevent-tasks',
    version=A['version'],
    author=A['author'],
    author_email=A['contact'],
    url=A['url'],
    description='Background task management using gevent and green threads.',
    long_description=long_description,
    packages=find_packages(),
    include_package_data=True,
    platforms=['any'],
    install_requires=list(requirements),
    cmdclass={
        'test': RunTests
    },
    classifiers=[
        'Topic :: Software Development :: Libraries',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Development Status :: 4 - Beta',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6'
    ]
)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# >>
#     gevent-tasks, 2017
# <<

import time
from typing import List, Generator
from numbers import Real
from logging import getLogger

import gevent
from gevent import Greenlet, Timeout
from gevent.pool import Pool


class TaskPool(Pool):
    """ Custom gevent thread pool for reporting capacity statistics. """

    def __init__(self, size=None):
        # type: (int) -> self
        if size is not None:
            size = max(2, size)
        super(TaskPool, self).__init__(size, Greenlet)

    def __repr__(self):
        return '<TaskPool(size=%d,running=%d,capacity=%0.1f%%)>' % (
            self.size, self.running, self.capacity)

    @property
    def running(self):
        # type: () -> int
        """ The number of currently running tasks in our pool. """
        return self.size - self.free_count()

    @property
    def capacity(self):
        # type: () -> float
        """ The current run capacity of our pool; how full as a percentage. """
        return (1 - (self.free_count() / float(self.size))) * 100


class Timing(object):
    """ Instance inside of Task object tracks running times of that task. """

    __slots__ = ('_first_start', '_run_times', '_started', 'name')

    def __init__(self, task_name=None):
        # type: (str) -> self
        self._first_start = 0   # type: float
        self._started = 0       # type: float
        self._run_times = []    # type: List[float]
        self.name = task_name or ''

    def __iter__(self):
        return self.history()

    def __repr__(self):
        return '<Timing(%scount=%d,started=%0.2f,last=%0.4f)>' % (
            ('name=%s,' % self.name) if self.name else '', self.count,
            self._first_start, self.last)

    def log(self, timing):
        # type: (float) -> None
        """ Mark a new finished time for the current task. """
        self._run_times.append(timing)

    def start(self):
        # type: () -> None
        """ Mark a new start time for the current task. """
        self._started = time.time()
        if self._first_start == 0:
            self._first_start = self._started

    def history(self):
        # type: () -> Generator[float]
        """ Iterator over all the saved timings. """
        for o in self.run_timings:
            yield o

    @property
    def started(self):
        # type: () -> float
        """ Unix timestamp of the last/currently running task started. """
        return self._started

    @property
    def first_started(self):
        # type: () -> float
        """ Unix timestamp of the first run of the task started. """
        return self._first_start

    @property
    def run_timings(self):
        # type: () -> List[float]
        """ Read-only getting for the collection of run timings.  """
        if not self._run_times:
            return [-1]
        return self._run_times

    @property
    def average(self):
        """ Total runtime divided by the successful runs count. """
        # type: () -> float
        return self.total / max(1, self.count)

    @property
    def count(self):
        """ Total successful runs of a given task. """
        # type: () -> int
        return len(self.run_timings)

    @property
    def last(self):
        # type: () -> float
        """ Runtime of the last task call. """
        return self.run_timings[-1]

    @property
    def total(self):
        # type: () -> float
        """ Total runtime, so far, of all task runs. """
        return sum(self.run_timings)

    @property
    def best(self):
        # type: () -> float
        """ Shortest single runtime for the task. """
        return min(self.run_timings)

    @property
    def worst(self):
        # type: () -> float
        """ Longest single runtime for the task. """
        return max(self.run_timings)


class Task(object):
    def __init__(self, name, fn, args=None, kwargs=None, timeout=None,
                 interval=None, description=None, logger=None, pool=None):

        if timeout is None:
            timeout = -1

        if args is None:
            args = tuple()

        if kwargs is None:
            kwargs = {}

        self.name = name                # type: str
        self.description = description  # type: str
        self.logger = logger or getLogger('%s.Task.%s' % (__name__, self.name))
        # ~~
        self._fn = fn                    # type: Callable
        self._fn_arg = args              # type: tuple
        self._fn_kw = kwargs             # type: dict
        self._g = None                   # type: Greenlet
        self._running = False            # type: bool
        self._schedule = False           # type: bool
        self._timeout_secs = timeout     # type: float
        self._timeout_obj = None         # type: Timeout
        self._interval = interval        # type: float
        self.timing = Timing(self.name)  # type: Timing
        self.pool = pool                 # type: TaskPool

    def __repr__(self):
        return '<Task(name=%s)>' % self.name

    def __make(self):
        # type: () -> Greenlet
        g = Greenlet(self._fn, self, *self._fn_arg, **self._fn_kw)
        g.link(self.__callback)
        return g

    def __callback(self, *args):
        duration = time.time() - self.timing.started
        if self._timeout_secs and self._timeout_obj and \
                self._timeout_obj.pending:
            self.logger.debug('canceling timeout')
            self._timeout_obj.cancel()
        self._g = None
        self._running = False
        self.timing.log(duration)
        if self.is_periodic and self._schedule:
            when = max(0, self._interval - duration)
            gevent.spawn_later(when, self.start)
            self.logger.debug('scheduled to run in %0.2f', when)

    def start(self):
        # type: () -> None
        if self.is_running:
            self.logger.warning('task is already running')
        elif self._g:
            self.logger.error('task is already claimed greenlet')
        else:
            self._schedule = True

        if not self._schedule:
            return

        if self._timeout_obj:
            self._timeout_obj.cancel()
            self._timeout_obj = None

        if self._timeout_secs > 0:
            self._timeout_obj = Timeout.start_new(
                self._timeout_secs, TimeoutError)

        self.timing.start()

        try:
            self._g = self.__make()
            self.pool.start(self._g)
            self._running = True

        except TimeoutError as e:
            self.logger.warning(e)
            self.__callback()

        except Exception as e:
            self.logger.exception(e, exc_info=True)
            raise e

    def stop(self, force=False):
        # type: (bool) -> None
        if self.is_running and self._g is not None:
            self._g.unlink(self.__callback)
            if force:
                self._g.kill(block=True, timeout=self._timeout_secs)
        self._schedule = False
        self.__callback()

    @property
    def is_running(self):
        # type: () -> bool
        if self._interval:
            return self.timing.started + self._interval > time.time()
        return self._running  # guess

    @property
    def is_periodic(self):
        # type: () -> bool
        return isinstance(self._interval, Real)



